/*
 *
 * $Id: postgre.c,v 1.30 2004/05/09 23:13:40 friedl Exp $
 *
 * postgresql backend to ipac-ng
 * Copyright (C) 2001-2003 Al Zakharov
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * The author can be reached via email: kaiser13@mail2000.ru, or by 
 * fido: Al_Zaharov@p88.f58.n5005.z2.fidonet.org
 *
*/

#include "ipac.h"
#include <stdio.h>
#include <string.h>
#include <dirent.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/param.h>
#include <unistd.h>
#include <libpq-fe.h>
#include <time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdarg.h>
#include <fcntl.h>
#include <sys/stat.h>

extern char *ahost;

static char *pgoptions = NULL;
static char *pgtty = NULL;

#ifdef DEBUG_DB
#define DEBUG_DB_LOGFILE "dblog"
static int fd;
void debuglog(const char *format, ...);
#define DPRINTF(fmt, args...) debuglog(fmt, ##args)
#else
#define DPRINTF(fmt, args...)
#endif

static PGconn *conn;
static PGresult *res;

/* the statical variables used for returning results */
static data_record_type  *_timestamp_lst = NULL;
static int               _timestamp_max  = 0;
static rule_type         *_rules_lst     = NULL;
static int               _rules_max      = 0;

/* a safety margin of 0.5 percent for memory allocations */
#define MEMORY_SAFETY_MARGIN(x) (int)(x * 0.005 + 0.5)

/* plain file ipac interface entries */
int postgre_stor_open (int flag);
int postgre_stor_store_record (const data_record_type *data);
int postgre_stor_list_timestamps (timestamp_t start, timestamp_t end,
		timestamp_t **data, timestamp_t *just_before,
		timestamp_t *just_after, char *ahost);
int postgre_stor_get_records (timestamp_t timestamp_b, timestamp_t timestamp_e, 
		data_record_type **data, char *filter);
int postgre_stor_get_summary (timestamp_t timestamp_b, timestamp_t timestamp_e,
		data_record_type **data, char *filter);
int postgre_stor_delete_record (timestamp_t timestamp);
void postgre_stor_close ();
void postgre_stor_clear ();

static const storage_method_t interface_entry = {
	"postgre",
	postgre_stor_open,
	postgre_stor_store_record,
	postgre_stor_list_timestamps,
	postgre_stor_get_records,
	postgre_stor_get_summary,
	postgre_stor_delete_record,
	postgre_stor_close
};

const storage_method_t *ipac_sm_interface_postgre () {
	return &interface_entry;
}

int postgre_stor_open (int flag)
{
#ifdef DEBUG_DB
	char *log_file;

	// construct the name of the logfile
	log_file = xmalloc(strlen(datadir) + sizeof(DEBUG_DB_LOGFILE) + 2);
	sprintf(log_file, "%s/" DEBUG_DB_LOGFILE, datadir);

	// open the logfile for appending
	fd = open (log_file, O_CREAT|O_WRONLY|O_APPEND|O_SYNC);
#endif	

	conn = PQsetdbLogin (dbhost, dbport, pgoptions, pgtty, dbname, dbuser,
									dbpass);
	if (PQstatus (conn) == CONNECTION_BAD) {
		fprintf (stderr, "Connection to database '%s' failed.\n", dbname);
		fprintf (stderr, "%s", PQerrorMessage (conn));

		DPRINTF ("failed postgre_stor_open\n");

		PQfinish (conn);
		return (1);
	}

	DPRINTF ("postgre_stor_open\n");
	storage_opened = 1;
	return (0);
}

int postgre_stor_store_record (const data_record_type *data)
{
	rule_type *rule, *firstrule;
	char wh_exec[380];

	firstrule = data->firstrule;

	if (firstrule == NULL)
		return 0;

	for (rule = firstrule; rule; rule = rule->next) {
		sprintf (wh_exec, "INSERT INTO logs (rule_name, bytes, pkts, that_time, hostname) "
				"VALUES ('%s', '%llu', '%llu', '%lu', '%s')", 
				rule->name, rule->bytes, rule->pkts, data->timestamp, hostname);

		DPRINTF ("postgre_stor_store_record\n");
		DPRINTF ("%s\n", wh_exec);

		res = PQexec (conn, wh_exec);
		if (!res || PQresultStatus (res) != PGRES_COMMAND_OK) {
			fprintf (stderr, "%s : %s\n", me, PQresultErrorMessage (res));
		}
		PQclear (res);
	}
	return 0;
}

int postgre_stor_list_timestamps (timestamp_t start, timestamp_t end,
		timestamp_t **data, timestamp_t *just_before,
		timestamp_t *just_after, char *ahost)
{
	int i;	
	timestamp_t *ts_list = NULL;
	int ts_list_len = 0;
	char wh_exec[250];

	DPRINTF ("postgre_stor_list_timestamps (%lu, %lu, **data, %s, %s, %s)\n", 
		start, end, (just_before!=NULL) ? "*just_before" : "NULL", 
		(just_after!=NULL) ? "*just_after" : "NULL", (ahost!=NULL) ? ahost : "NULL");

	sprintf (wh_exec, "SELECT distinct that_time FROM logs WHERE ");
	if (ahost)
		sprintf (wh_exec+strlen (wh_exec), "hostname = '%s' AND ", ahost);
	sprintf (wh_exec+strlen (wh_exec), "that_time between %lu and %lu "
		"ORDER BY that_time ASC", start, end);

	DPRINTF ("%s\n", wh_exec);

	res = PQexec (conn, wh_exec);
	if (!res || PQresultStatus (res) != PGRES_TUPLES_OK) {
		fprintf (stderr, "%s: %s\n", me, PQresultErrorMessage (res));

		DPRINTF ("%s: %s\n", me, PQresultErrorMessage (res));

		return (-1);
	}
	ts_list_len = PQntuples (res);

	if (ts_list_len > 0) {
		ts_list = xmalloc (PQntuples (res) * sizeof (timestamp_t));
		for (i = 0; i < ts_list_len; i++)
			ts_list[i] = strtoul ((char *) PQgetvalue (res, i, 0), NULL, 0);
		PQclear (res);

		if (just_before != NULL) {
			sprintf (wh_exec, "SELECT that_time FROM logs WHERE "
					"that_time < %lu ORDER BY that_time DESC LIMIT 1", start);

			DPRINTF ("%s\n", wh_exec);

			res = PQexec (conn, wh_exec);
			if (!res || PQresultStatus (res) != PGRES_TUPLES_OK) {
				fprintf (stderr, "%s: %s\n", me, PQresultErrorMessage (res));

				DPRINTF ("%s: %s\n", me, PQresultErrorMessage (res));

				if (ts_list) free (ts_list);
				return (-1);
			} 
			if (PQntuples (res) > 0)
				*just_before = strtoul ((char *) PQgetvalue (res, 0, 0), NULL, 0);
			else
				*just_before = (timestamp_t) -1 ;
			PQclear (res);
		}

		if (just_after != NULL) {
			sprintf (wh_exec, "SELECT that_time FROM logs WHERE "
					"that_time > %lu ORDER BY that_time ASC LIMIT 1", end);

			DPRINTF ("%s\n", wh_exec);

			res = PQexec (conn, wh_exec);
			if (!res || PQresultStatus (res) != PGRES_TUPLES_OK) {
				fprintf (stderr, "%s: %s\n", me, PQresultErrorMessage (res));

				DPRINTF ("%s: %s\n", me, PQresultErrorMessage (res));

				if (ts_list) free (ts_list);
				return (-1);
			} 
			if (PQntuples (res) > 0)
				*just_after = strtoul ((char *) PQgetvalue (res, 0, 0), NULL, 0);
			else
				*just_after = (timestamp_t) -1;
			PQclear (res);
		}

		*data = ts_list;

	}

	DPRINTF ("postgre_stor_list_timestamps returning: %i\n", ts_list_len);
	return ts_list_len;
}

int postgre_stor_get_records (timestamp_t timestamp_b, timestamp_t timestamp_e, 
		data_record_type **data, char *filter)
{
	rule_type *r, *r1 = NULL, *rules;
	int i, nr_timestamps, index, nr_rules;
	timestamp_t timestamp_akt;
	char wh_exec[320];
	/* char *tmp; */

	DPRINTF ("postgre_stor_get_records (%lu, %lu, **data, %s)\n", 
		timestamp_b, timestamp_e, (filter!=NULL) ? filter : "NULL");

	if (timestamp_e) {
		sprintf (wh_exec, "SELECT count (distinct that_time) FROM logs WHERE ");
		if (ahost)
			sprintf (wh_exec+strlen (wh_exec), "hostname = '%s' AND ", ahost);
		sprintf (wh_exec+strlen (wh_exec), "that_time between '%lu' and '%lu'", 
						 timestamp_b, timestamp_e);
		if (filter)
			sprintf (wh_exec+strlen (wh_exec), " AND rule_name like '%%%s%%'", 
							filter);

		DPRINTF ("%s\n", wh_exec);

		res = PQexec (conn, wh_exec);
		if (!res || PQresultStatus (res) != PGRES_TUPLES_OK) {
			fprintf (stderr, "%s: PQexec of (%s) command didn't return tuples"
					" properly\nlibpq says: %s", me, wh_exec, 
					PQresultErrorMessage (res));

			DPRINTF ("%s\n", PQresultErrorMessage (res));

			PQclear (res);
			PQfinish(conn);
			exit (-1);
		}
		nr_timestamps = strtol (PQgetvalue (res,0,0),0,0);
		DPRINTF ("result: %u timestamps\n", nr_timestamps);
		PQclear (res);
	} else {
		nr_timestamps = 1;
		DPRINTF ("searching for exactly one timestamp: %u\n", timestamp_b);
	}
	
	sprintf (wh_exec, "SELECT rule_name, bytes, pkts, that_time, hostname FROM logs WHERE ");
	if (ahost)
		sprintf (wh_exec+strlen (wh_exec), "hostname = '%s' AND ", ahost);
	if (timestamp_e) {
		sprintf (wh_exec+strlen (wh_exec), "that_time between '%lu' and '%lu'",
						timestamp_b, timestamp_e);
	} else {
		sprintf (wh_exec+strlen (wh_exec), "that_time = '%lu'", timestamp_b);
	}
	if (filter)
		sprintf (wh_exec+strlen (wh_exec), " AND rule_name like '%%%s%%'", filter);
	if (timestamp_e)
		sprintf (wh_exec+strlen (wh_exec), " ORDER BY that_time");

	DPRINTF ("%s\n", wh_exec);

	res = PQexec (conn, wh_exec);
	if (!res || PQresultStatus (res) != PGRES_TUPLES_OK) {
		fprintf (stderr, "%s: PQexec of (%s) command didn't return tuples"
			" properly\nlibpq says: %s", me, wh_exec,
			PQresultErrorMessage (res));

		DPRINTF ("%s\n", PQresultErrorMessage (res));

		PQclear (res);
		PQfinish(conn);
		exit (-1);
	}
	nr_rules = PQntuples (res);
	DPRINTF ("result: %u data records\n", nr_rules);

	DPRINTF ("starting to convert data into our format\n");

	/* create record_data_type. */
	if (nr_timestamps > _timestamp_max) {
		/* need to increase size of data_record_type array */

		if (_timestamp_lst != NULL) {
			/* as realloc does not support a count parameter we free the old list first 
			   and then call calloc again */
			DPRINTF ("freeing memory of old timestamp list of size %i\n", _timestamp_max);
			free (_timestamp_lst);
		}

		/* always add a safety margin */
		_timestamp_max = nr_timestamps + MEMORY_SAFETY_MARGIN(nr_timestamps);

		DPRINTF ("allocating memory for timestamp list with %i elements\n", _timestamp_max);
		DPRINTF ("  calloc (cnt %i, size %u) = %lu bytes\n", _timestamp_max, sizeof (data_record_type), (long)((long)_timestamp_max * (long)sizeof (data_record_type)));
		_timestamp_lst = (data_record_type *)calloc (_timestamp_max, sizeof (data_record_type));
	
		if (_timestamp_lst == NULL) {
			fprintf (stderr,"%s: data_record_type calloc(cnt %i, size %u) failed: %s\n", me, _timestamp_max, sizeof (data_record_type), strerror(errno));

			DPRINTF ("failed: %s\n", strerror(errno));

			PQclear (res);
			PQfinish(conn);
			exit (-1);
		}
	}
	*data = _timestamp_lst;

	/* create rule_type. */
	if (nr_rules > _rules_max) {
		/* need to increase size of rule_type array */
		
		if (_rules_lst != NULL) {
			/* as realloc does not support a count parameter we free the old list first 
			   and then call calloc again */
			DPRINTF ("freeing memory of old timestamp list of size %i\n", _rules_max);
			free (_rules_lst);
		}

		/* always add a safety margin */
		_rules_max = nr_rules + MEMORY_SAFETY_MARGIN(nr_rules);

		DPRINTF ("allocating memory for rule list with %i elements\n", _rules_max);
		DPRINTF ("  calloc (cnt %i, size %u) = %lu bytes\n", _rules_max, sizeof (rule_type), (long)((long)_rules_max * (long)sizeof (rule_type)));
		_rules_lst = (rule_type *)calloc (_rules_max, sizeof (rule_type));
	
		if (_rules_lst == NULL) {
			fprintf (stderr,"%s: rule_type calloc(cnt %i, size %u) failed: %s\n", me, _rules_max, sizeof (rule_type), strerror(errno));

			DPRINTF ("failed: %s\n", strerror(errno));

			PQclear (res);
			PQfinish(conn);
			exit (-1);
		}
	}
	rules = _rules_lst;

	index = -1;
	timestamp_akt = 0;
	for (i = 0; i < nr_rules; i++) {
		// currently timestamp_t is of type time_t which is essentially long
		timestamp_t tstamp_new = strtol ((char *) PQgetvalue (res,i,3), NULL, 10);
		if (tstamp_new != timestamp_akt) { // do we have a new timestamp?
			timestamp_akt = tstamp_new;
			index++;
			if (index > nr_timestamps) {
				fprintf (stderr,"%s: We got more records than timestamps "
					"were reported before. This should not happen!\n", me);

				DPRINTF ("We got more records than timestamps "
					"were reported before. This should not happen!\n");

				PQclear (res);
				PQfinish(conn);
				exit (-1);
			}
			(*data)[index].timestamp = timestamp_akt;
			/* FIXME: Two records with the same 'that_time' field but different 
			          hostname fields get listed as being from the hostname of 
			          the first record. Therefore it makes no sense and is plain 
			          wrong to store the hostname in this query. */
			
			/* tmp = (char *)PQgetvalue (res, i, 4);
			(*data)[index].machine_name = calloc(1, strlen(tmp)+1);
			
			if ((*data)[index].machine_name == NULL) {
				fprintf (stderr,"%s: calloc(1, size %u) for string \"%s\" failed: %s\n", me, strlen(tmp)+1, tmp, strerror(errno));

				DPRINTF ("calloc(1, size %u) for string \"%s\" failed: %s\n", strlen(tmp)+1, tmp, strerror(errno));

				PQclear (res);
				PQfinish(conn);
				exit (-1);
			}
			memcpy((*data)[index].machine_name, tmp, strlen(tmp)+1); */
			(*data)[index].machine_name = NULL;
			
			(*data)[index].firstrule = NULL;
			r1 = NULL;
		}

		r = &rules[i];
		r->next = NULL;

		if (r1 == NULL)
			 (*data)[index].firstrule = r;
		else
			r1->next = r;
		r1 = r;

		// never copy more than MAX_RULE_NAME_LENGTH+1 bytes from the resulting rule name
		strncpy (r->name, (char *) PQgetvalue (res, i, 0), MAX_RULE_NAME_LENGTH+1);
		// make sure it's a null terminated string
		r->name[MAX_RULE_NAME_LENGTH] = '\0';
		r->bytes = strtoull ((char *) PQgetvalue (res, i, 1), NULL, 10);
		r->pkts = strtoull ((char *) PQgetvalue (res, i, 2), NULL, 10);

#ifdef DEBUG_DB_LEVEL2
		DPRINTF ("Record: %s, bytes %llu, pkts %llu\n", r->name, r->bytes, r->pkts);
#endif	
	}
	PQclear (res);

	DPRINTF ("data conversion finished\n");
	DPRINTF ("postgre_stor_get_records returning: %i\n", index+1);
	return index+1;
}

int postgre_stor_get_summary (timestamp_t timestamp_b, timestamp_t timestamp_e, 
		data_record_type **data, char *filter)
{
	rule_type *r, *r1 = NULL;
	int i, nr_timestamps = 0, index;
	timestamp_t timestamp_akt;
	char wh_exec[320];

	DPRINTF ("postgre_stor_get_summary (%lu, %lu, **data, %s)\n", 
		timestamp_b, timestamp_e, (filter!=NULL) ? filter : "NULL");

	if (!timestamp_e) 
		timestamp_e = timestamp_b;

	if (ahost)
		sprintf (wh_exec, "SELECT rule_name, sum (bytes), sum (pkts), hostname FROM logs "
			"WHERE "
			"hostname = '%s' and that_time between '%lu' and '%lu' ",
			 ahost, timestamp_b, timestamp_e);
	else
		sprintf (wh_exec, "SELECT rule_name, sum (bytes), sum (pkts), hostname FROM logs "
				"WHERE "
				"that_time between '%lu' and '%lu' ",
				timestamp_b, timestamp_e);
	if (filter)
		sprintf (wh_exec+strlen (wh_exec), " and rule_name like '%%%s%%'", filter);

	sprintf (wh_exec+strlen (wh_exec), " group by rule_name, hostname");

	DPRINTF ("%s\n", wh_exec);

	res = PQexec (conn, wh_exec);
	if (!res || PQresultStatus (res) != PGRES_TUPLES_OK) {
		fprintf (stderr, "PQexec of (%s) command didn't return tuples"
			" properly\nlibpq says: %s", wh_exec,
			PQresultErrorMessage (res));

		DPRINTF ("%s: %s\n", me, PQresultErrorMessage (res));

		PQclear (res);
		PQfinish(conn);
		exit (-1);
	}

	DPRINTF ("starting to convert data into our format\n");

	/* create record_data_type. */
	*data = (data_record_type *)xmalloc (sizeof (data_record_type));

	index = -1;
	timestamp_akt = 0;
	for (i = 0; i < PQntuples (res); i++) {
		timestamp_t tstamp_new = timestamp_b; 
		if (tstamp_new != timestamp_akt) { // do we have a new timestamp?
			timestamp_akt = tstamp_new;
			index++;
			if (index > nr_timestamps) {
				fprintf (stderr,"We got more records then timestamps "
					"were reported before. This should not happen\n");

				DPRINTF ("We got more records then timestamps "
					"were reported before. This should not happen\n");

				PQclear (res);
				PQfinish(conn);
				exit (-1);
			}
			(*data)[index].timestamp = timestamp_akt;
			(*data)[index].machine_name = xstrdup ((char *) PQgetvalue (res, i, 3));
			(*data)[index].firstrule = NULL;
			r1 = NULL;
		}
		r = new_rule ();
		if (r1 == NULL)
			 (*data)[index].firstrule = r;
		else
			r1->next = r;
		r1 = r;

		// never copy more than MAX_RULE_NAME_LENGTH+1 bytes from the resulting rule name
		strncpy (r->name, (char *) PQgetvalue (res, i, 0), MAX_RULE_NAME_LENGTH+1);
		// make sure it's a null terminated string
		r->name[MAX_RULE_NAME_LENGTH] = '\0';
		r->bytes = strtoull ((char *) PQgetvalue (res, i, 1), NULL, 10);
		r->pkts = strtoull ((char *) PQgetvalue (res, i, 2), NULL, 10);
	}
	PQclear (res);

	DPRINTF ("data conversion finished\n");
	DPRINTF ("postgre_stor_get_summary returning: %i\n", index+1);
	return index+1;
}

int postgre_stor_delete_record (timestamp_t timestamp)
{
	char wh_exec[120];

	sprintf (wh_exec, "DELETE FROM logs WHERE that_time = '%lu'", timestamp);

	DPRINTF ("postgre_stor_delete_record\n");
	DPRINTF ("%s\n", wh_exec);

	res = PQexec (conn, wh_exec);
	PQclear (res);
	return 0;
}

void postgre_stor_close ()
{
	DPRINTF ("postgre_stor_close()\n");

	/* clean up first */
	postgre_stor_clear ();

	PQfinish (conn);
	storage_opened = 0;
	DPRINTF ("postgre_stor_close finished\n");

#ifdef DEBUG_DB
	/* close logfile */
	if (fd != -1) {
		close (fd);
	}
#endif

}

/* free data we kept in memory */
void postgre_stor_clear ()
{
	DPRINTF ("postgre_stor_clear() \n");

	if (_timestamp_lst != NULL) {
		DPRINTF ("freeing data_record_type array with %i elements\n", _timestamp_max);
		free_data_record_type_array (_timestamp_lst, _timestamp_max);
		_timestamp_lst = NULL;
		_timestamp_max = 0;
	}
	
	if (_rules_lst != NULL) {
		DPRINTF ("freeing rule_type array with %i elements\n", _rules_max);
		free (_rules_lst);
		_rules_lst = NULL;
		_rules_max = 0;
	}
	DPRINTF ("postgre_stor_clear finished\n");
}

#ifdef DEBUG_DB
void debuglog (const char *format, ...)
{
	va_list arg;
	time_t tm;
	char logline[512];
	
	// construct the date string like syslog does, eg. Sep 26 00:23:04
	tm = time(NULL);	
	strftime(logline, sizeof(logline)-1, "%b %e %H:%M:%S ", localtime(&tm));

	snprintf (logline+strlen(logline), sizeof(logline)-strlen(logline)-1, "fetchipac[%i]: ", getpid());
	    
	va_start (arg, format);
	vsnprintf(logline+strlen(logline), sizeof(logline)-strlen(logline)-1, format, arg);
	va_end (arg);

	// write the log message to the logfile
	if (fd != -1) {
		write (fd, logline, strlen(logline));
	}
}
#endif	
